@@ -1,11 +1,11 @@
 use strict;
 use warnings;
-package JSON::Schema::Modern; # git description: v0.598-8-ga3c37e8e
+package JSON::Schema::Modern; # git description: v0.599-24-ged06ea98
 # vim: set ts=8 sts=2 sw=2 tw=100 et :
 # ABSTRACT: Validate data against a schema using a JSON Schema
 # KEYWORDS: JSON Schema validator data validation structure specification
 
-our $VERSION = '0.599';
+our $VERSION = '0.600';
 
 use 5.020;  # for fc, unicode_strings features
 use Moo;
@@ -110,8 +110,6 @@ has _format_validations => (
       sub => CodeRef,
     ]],
   init_arg => 'format_validations',
-  lazy => 1,
-  default => sub { {} },
 );
 
 sub _get_format_validation ($self, $format) { ($self->{_format_validations}//{})->{$format} }
@@ -159,13 +157,28 @@ sub add_schema {
   Carp::carp('use of deprecated form of add_schema with document')
     if $_[0]->$_isa('JSON::Schema::Modern::Document');
 
+  return $self->add_document($uri, $_[0]) if $_[0]->$_isa('JSON::Schema::Modern::Document');
+
   # document BUILD will trigger $self->traverse($schema)
-  my $document = $_[0]->$_isa('JSON::Schema::Modern::Document') ? shift
-    : JSON::Schema::Modern::Document->new(
-      schema => shift,
-      $uri ? (canonical_uri => $uri) : (),
-      evaluator => $self,  # used mainly for traversal during document construction
-    );
+  # Note we do not pass the uri to the document constructor, so resources in that document may still
+  # be relative
+  my $document = JSON::Schema::Modern::Document->new(
+    schema => $_[0],
+    evaluator => $self,  # used mainly for traversal during document construction
+  );
+
+  # try to reuse the same document, if the same schema is being added twice:
+  # this results in _add_resource silently ignoring the duplicate add, rather than erroring.
+  my $schema_checksum = $document->_checksum(md5($self->_json_decoder->encode($document->schema)));
+  if (my $existing_doc = first {
+        my $existing_checksum = $_->_checksum
+          // $_->_checksum(md5($self->_json_decoder->encode($_->schema)));
+        $existing_checksum eq $schema_checksum
+          and $_->canonical_uri eq $document->canonical_uri
+          # FIXME: must also check spec version/metaschema_uri/vocabularies
+      } uniqint map $_->{document}, $self->_canonical_resources) {
+    $document = $existing_doc;
+  }
 
   $self->add_document($uri, $document);
 }
@@ -175,10 +188,10 @@ sub add_document {
   my $self = shift;
 
   # TODO: resolve $uri against $self->base_uri
-  my $uri = !is_ref($_[0]) ? Mojo::URL->new(shift)
+  my $base_uri = !is_ref($_[0]) ? Mojo::URL->new(shift)
     : $_[0]->$_isa('Mojo::URL') ? shift : Mojo::URL->new;
 
-  croak 'cannot add a schema with a uri with a fragment' if defined $uri->fragment;
+  croak 'cannot add a schema with a uri with a fragment' if defined $base_uri->fragment;
 
   my $document = shift or croak 'insufficient arguments';
   croak 'wrong document type' if not $document->$_isa('JSON::Schema::Modern::Document');
@@ -190,35 +203,39 @@ sub add_document {
     exception => 1,
   ) if $document->has_errors;
 
-  if (not grep refaddr($_->{document}) == refaddr($document), $self->_canonical_resources) {
-    my $schema_checksum = $document->_checksum
-      // $document->_checksum(md5($self->_json_decoder->encode($document->schema)));
+  my @root; # uri_string => resource hash of the resource at path ''
 
-    if (my $existing_doc = first {
-          my $existing_checksum = $_->_checksum
-            // $_->_checksum(md5($self->_json_decoder->encode($_->schema)));
-          $existing_checksum eq $schema_checksum
-        } uniqint map $_->{document}, $self->_canonical_resources) {
-      # we already have this schema content in another document object.
-      $document = $existing_doc;
-    }
-    else {
-      $self->_add_resources(map +($_->[0] => +{ $_->[1]->%*, document => $document }),
-        $document->resource_pairs);
+  # document resources are added after resolving each resource against our provided base uri
+  foreach my $res_pair ($document->resource_pairs) {
+    my ($uri_string, $doc_resource) = @$res_pair;
+    $uri_string = Mojo::URL->new($uri_string)->to_abs($base_uri)->to_string if length $base_uri;
+
+    my $new_resource = {
+      $doc_resource->%{qw(path specification_version vocabularies configs)},
+      document => $document,
+    };
+
+    $new_resource->{canonical_uri} = length $base_uri
+      ? Mojo::URL->new($doc_resource->{canonical_uri})->to_abs($base_uri)
+      : $doc_resource->{canonical_uri};
+
+    foreach my $anchor (keys (($doc_resource->{anchors}//{})->%*)) {
+      use autovivification 'store';
+      $new_resource->{anchors}{$anchor} = {
+        path => $doc_resource->{anchors}{$anchor}{path},
+        canonical_uri => length $base_uri
+          ? Mojo::URL->new($doc_resource->{anchors}{$anchor}{canonical_uri})->to_abs($base_uri)
+          : $doc_resource->{anchors}{$anchor}{canonical_uri},
+      };
     }
+
+    # this might croak if there are duplicates or malformed entries.
+    $self->_add_resource($uri_string => $new_resource);
+    @root = ( $uri_string => $new_resource ) if $new_resource->{path} eq '' and $uri_string !~ /#./;
   }
 
-  if ("$uri") {
-    my $resource = $document->_get_resource($document->canonical_uri);
-    $self->_add_resources($uri => {
-        path => '',
-        canonical_uri => $document->canonical_uri,
-        specification_version => $resource->{specification_version},
-        vocabularies => $resource->{vocabularies},  # reference, not copy
-        document => $document,
-        configs => $resource->{configs},
-      });
-  }
+  # associate the root resource with the base uri we were provided, if it does not already exist
+  $self->_add_resource($base_uri.'' => $root[1]) if length $base_uri and $root[0] ne $base_uri;
 
   return $document;
 }
@@ -238,6 +255,7 @@ sub evaluate_json_string ($self, $json_data, $schema, $config_override = {}) {
       errors => [
         JSON::Schema::Modern::Error->new(
           depth => 0,
+          mode => 'evaluate',
           keyword => undef,
           instance_location => '',
           keyword_location => '',
@@ -257,14 +275,23 @@ sub evaluate_json_string ($self, $json_data, $schema, $config_override = {}) {
 # Returns the internal $state object accumulated during the traversal.
 sub traverse ($self, $schema_reference, $config_override = {}) {
   my %overrides = %$config_override;
-  delete @overrides{qw(callbacks initial_schema_uri metaschema_uri traversed_schema_path)};
+  delete @overrides{qw(callbacks initial_schema_uri metaschema_uri traversed_schema_path specification_version)};
   croak join(', ', sort keys %overrides), ' not supported as a config override in traverse'
     if keys %overrides;
 
   # Note: the starting position is not guaranteed to be at the root of the $document.
   my $initial_uri = Mojo::URL->new($config_override->{initial_schema_uri} // '');
   my $initial_path = $config_override->{traversed_schema_path} // '';
-  my $spec_version = $self->specification_version//SPECIFICATION_VERSION_DEFAULT;
+  my $spec_version = $config_override->{specification_version} // $self->specification_version // SPECIFICATION_VERSION_DEFAULT;
+
+  croak 'traversed_schema_path must be a json pointer' if $initial_path !~ m{^(?:/|$)};
+
+  if (length(my $uri_path = $initial_uri->fragment)) {
+    croak 'initial_schema_uri fragment must be a json pointer' if $uri_path !~ m{^/};
+
+    croak 'traversed_schema_path does not match initial_schema_uri path fragment'
+      if $initial_path !~ m{\Q$uri_path\E$};
+  }
 
   my $state = {
     depth => 0,
@@ -274,7 +301,7 @@ sub traverse ($self, $schema_reference, $config_override = {}) {
     schema_path => '',                      # the rest of the path, since the start of this method or last $id
     effective_base_uri => Mojo::URL->new(''),
     errors => [],
-    identifiers => [],
+    identifiers => {},
     subschemas => [],
     configs => {},
     callbacks => $config_override->{callbacks} // {},
@@ -348,7 +375,7 @@ sub evaluate ($self, $data, $schema_reference, $config_override = {}) {
     }
     else {
       # traverse is called via add_schema -> ::Document->new -> ::Document->BUILD
-      my $document = $self->add_schema('', $schema_reference);
+      my $document = $self->add_schema($schema_reference);
       my $base_resource = $document->_get_resource($document->canonical_uri)
         || croak "couldn't get resource: document parse error";
 
@@ -742,40 +769,40 @@ sub _eval_subschema ($self, $data, $schema, $state) {
   return $valid;
 }
 
+my $path_type = Str->where('m{^(?:/|$)}');  # JSON pointer relative to the document root
 has _resource_index => (
   is => 'bare',
-  isa => HashRef[my $resource_type = Dict[
-      canonical_uri => InstanceOf['Mojo::URL'],
-      path => Str,
+  isa => Map[my $resource_key_type = Str->where('!/#/'), my $resource_type = Dict[
+      # always stringwise-equal to the top level key
+      canonical_uri => (InstanceOf['Mojo::URL'])->where(q{not defined $_->fragment}),
+      path => $path_type,
       specification_version => my $spec_version_type = Enum(SPECIFICATION_VERSIONS_SUPPORTED),
       document => InstanceOf['JSON::Schema::Modern::Document'],
       # the vocabularies used when evaluating instance data against schema
       vocabularies => ArrayRef[my $vocabulary_class_type = ClassName->where(q{$_->DOES('JSON::Schema::Modern::Vocabulary')})],
+      anchors => Optional[HashRef[Dict[
+        canonical_uri => (InstanceOf['Mojo::URL'])->where(q{not defined $_->fragment or substr($_->fragment, 0, 1) eq '/'}),
+        path => $path_type,
+      ]]],
       configs => HashRef,
       Slurpy[HashRef[Undef]],  # no other fields allowed
     ]],
-  lazy => 1,
-  default => sub { {} },
 );
 
 sub _get_resource { ($_[0]->{_resource_index}//{})->{$_[1]} }
-sub _add_resources {
-  use autovivification 'store';
-  $_[0]->{_resource_index}{$_->[0]} = $resource_type->($_->[1]) foreach pairs @_[1..$#_];
-}
+
+# does not check for duplicate entries, or for malformed uris
 sub _add_resources_unsafe {
   use autovivification 'store';
-  $_[0]->{_resource_index}{$_->[0]} = $resource_type->($_->[1]) foreach pairs @_[1..$#_];
+  $_[0]->{_resource_index}{$resource_key_type->($_->[0])} = $resource_type->($_->[1])
+    foreach pairs @_[1..$#_];
 }
-sub _resource_index { $_[0]->{_resource_index}->%* }
+sub _resource_index { ($_[0]->{_resource_index}//{})->%* }
 sub _canonical_resources { values(($_[0]->{_resource_index}//{})->%*) }
 sub _resource_pairs { pairs(($_[0]->{_resource_index}//{})->%*) }
 
-around _add_resources => sub {
-  my ($orig, $self) = (shift, shift);
-
-  my @resources;
-  foreach my $pair (sort { $a->[0] cmp $b->[0] } pairs @_) {
+sub _add_resource ($self, @kvs) {
+  foreach my $pair (sort { $a->[0] cmp $b->[0] } pairs @kvs) {
     my ($key, $value) = @$pair;
 
     if (my $existing = $self->_get_resource($key)) {
@@ -793,16 +820,10 @@ around _add_resources => sub {
       croak 'uri "'.$key.'" conflicts with an existing meta-schema resource';
     }
 
-    my $fragment = $value->{canonical_uri}->fragment;
-    croak sprintf('canonical_uri cannot contain an empty fragment (%s)', $value->{canonical_uri})
-      if defined $fragment and $fragment eq '';
-
-    croak sprintf('canonical_uri cannot contain a plain-name fragment (%s)', $value->{canonical_uri})
-      if ($fragment // '') =~ m{^[^/]};
-
-    $self->$orig($key, $value);
+    use autovivification 'store';
+    $self->{_resource_index}{$resource_key_type->($key)} = $resource_type->($value);
   }
-};
+}
 
 # $vocabulary uri (not its $id!) => [ spec_version, class ]
 has _vocabulary_classes => (
@@ -890,6 +911,7 @@ sub _get_metaschema_info ($self, $metaschema_uri, $for_canonical_uri) {
         my $absolute_location = $e->absolute_keyword_location // $for_canonical_uri;
         JSON::Schema::Modern::Error->new(
           depth => $e->depth,
+          mode => 'traverse',
           keyword => $e->keyword eq '$schema' ? '' : $e->keyword,
           instance_location => $e->instance_location,
           keyword_location => ($for_canonical_uri->fragment//'').($e->keyword_location =~ s{^/\$schema\b}{}r),
@@ -1018,55 +1040,51 @@ sub _get_or_load_resource ($self, $uri) {
 };
 
 # returns information necessary to use a schema found at a particular URI or uri-reference:
-# - a schema (which may not be at a document root)
-# - the canonical uri for that schema,
-# - the JSON::Schema::Modern::Document object that holds that schema
-# - the path relative to the document root for this schema
-# - the specification version that applies to this schema
-# - the vocabularies to use when considering schema keywords
-# - the config overrides to set when considering schema keywords
+# - schema: a schema (which may not be at a document root)
+# - canonical_uri: the canonical uri for that schema,
+# - document: the JSON::Schema::Modern::Document object that holds that schema
+# - document_path: the path relative to the document root for this schema
+# - specification_version: the specification version that applies to this schema
+# - vocabularies: the vocabularies to use when considering schema keywords
+# - configs: the config overrides to set when considering schema keywords
 # creates a Document and adds it to the resource index, if not already present.
 sub _fetch_from_uri ($self, $uri_reference) {
   $uri_reference = Mojo::URL->new($uri_reference) if not is_ref($uri_reference);
   my $fragment = $uri_reference->fragment;
 
-  if (not length($fragment) or $fragment =~ m{^/}) {
-    my $base = $uri_reference->clone->fragment(undef);
-    if (my $resource = $self->_get_or_load_resource($base)) {
-      my $subschema = $resource->{document}->get(my $document_path = $resource->{path}.($fragment//''));
-      return if not defined $subschema;
-      my $document = $resource->{document};
-      my $closest_resource = first { !length($_->[1]{path})       # document root
-          || length($document_path)
-            && $document_path =~ m{^\Q$_->[1]{path}\E(?:/|\z)} }  # path is above present location
-        sort { length($b->[1]{path}) <=> length($a->[1]{path}) }  # sort by length, descending
-        grep { not length Mojo::URL->new($_->[0])->fragment }     # omit anchors
-        $document->resource_pairs;
+  my $resource = $self->_get_or_load_resource($uri_reference->clone->fragment(undef));
+  return if not $resource;
 
-      my $canonical_uri = $closest_resource->[1]{canonical_uri}->clone
-        ->fragment(substr($document_path, length($closest_resource->[1]{path})));
-      $canonical_uri->fragment(undef) if not length($canonical_uri->fragment);
-      return {
-        schema => $subschema,
-        canonical_uri => $canonical_uri,
-        document => $document,
-        document_path => $document_path,
-        $resource->%{qw(specification_version vocabularies configs)}, # reference, not copy
-      };
-    }
+  if (not length($fragment) or $fragment =~ m{^/}) {
+    my $subschema = $resource->{document}->get(my $document_path = $resource->{path}.($fragment//''));
+    return if not defined $subschema;
+    my $doc_addr = refaddr($resource->{document});
+    my $closest_resource = first { !length($_->[1]{path})       # document root
+        || length($document_path)
+          && $document_path =~ m{^\Q$_->[1]{path}\E(?:/|\z)} }  # path is above present location
+      sort { length($b->[1]{path}) <=> length($a->[1]{path}) }  # sort by length, descending
+      grep { refaddr($_->[1]{document}) == $doc_addr }          # in same document
+      $self->_resource_pairs;
+
+    my $canonical_uri = $closest_resource->[1]{canonical_uri}->clone
+      ->fragment(substr($document_path, length($closest_resource->[1]{path})));
+    $canonical_uri->fragment(undef) if not length($canonical_uri->fragment);
+
+    return {
+      schema => $subschema,
+      canonical_uri => $canonical_uri,
+      document_path => $document_path,
+      $resource->%{qw(document specification_version vocabularies configs)}, # reference, not copy
+    };
   }
   else {  # we are following a URI with a plain-name fragment
-    if (my $resource = $self->_get_resource($uri_reference)) {
-      my $subschema = $resource->{document}->get($resource->{path});
-      return if not defined $subschema;
-      return {
-        schema => $subschema,
-        canonical_uri => $resource->{canonical_uri}->clone, # this is *not* the anchor-containing URI
-        document => $resource->{document},
-        document_path => $resource->{path},
-        $resource->%{qw(specification_version vocabularies configs)}, # reference, not copy
-      };
-    }
+    return if not my $subresource = ($resource->{anchors}//{})->{$fragment};
+    return {
+      schema => $resource->{document}->get($subresource->{path}),
+      canonical_uri => $subresource->{canonical_uri}->clone, # this is *not* the anchor-containing URI
+      document_path => $subresource->{path},
+      $resource->%{qw(document specification_version vocabularies configs)}, # reference, not copy
+    };
   }
 }
 
@@ -1194,7 +1212,7 @@ JSON::Schema::Modern - Validate data against a schema using a JSON Schema
 
 =head1 VERSION
 
-version 0.599
+version 0.600
 
 =head1 SYNOPSIS
 
@@ -1220,12 +1238,12 @@ These values are all passed as arguments to the constructor.
 
 =head2 specification_version
 
-Indicates which version of the JSON Schema specification is used during evaluation. When not set,
-this value is derived from the C<$schema> keyword in the schema used in evaluation, or defaults to
-the latest version (currently C<draft2020-12>).
+Indicates which version of the JSON Schema specification is used during evaluation. This value is
+overridden by the value determined from the C<$schema> keyword in the schema used in evaluation
+(when present), or defaults to the latest version (currently C<draft2020-12>).
 
-The use of this option is I<HIGHLY> encouraged to ensure continued correct operation of your schema.
-The current default value will not stay the same over time.
+The use of the C<$schema> keyword in your schema is I<HIGHLY> encouraged to ensure continued correct
+operation of your schema. The current default value will not stay the same over time.
 
 May be one of:
 
@@ -1253,9 +1271,6 @@ L<C<draft4> or C<4>|https://json-schema.org/specification-links.html#draft-4>, c
 
 =back
 
-Note that you can also use a C<$schema> keyword in the schema itself, to specify a different metaschema or
-specification version.
-
 =head2 output_format
 
 One of: C<flag>, C<basic>, C<strict_basic>, C<terse>. Defaults to C<basic>.
@@ -1597,13 +1612,17 @@ For example, to find the resolved targets of all C<$ref> keywords in a schema do
 
 Introduces the (unblessed, nested) Perl data structure
 representing a JSON Schema to the implementation, registering it under the indicated URI if
-provided, and all identifiers found within the document will be added as well (C<''> will be used if
-no other identifier can be found within).
+provided, and all identifiers found within the document will be resolved against this URI (if
+provided) and added as well. C<''> will be used if no other identifier can be found within.
 
 You B<MUST> call C<add_schema> or L</add_document> (below) for any external resources that a schema may reference via C<$ref>
 before calling L</evaluate>, other than the standard metaschemas which are loaded from a local cache
 as needed.
 
+If you add multiple schemas (either with this method, or implicitly via L</evaluate>) with no root
+identifier (either provided explicitly in the method call, or via an C<$id> keyword at the schema
+root), all such previous schemas are removed from memory and can no longer be referenced.
+
 If there were errors in the document, will die with these errors;
 otherwise returns the L<JSON::Schema::Modern::Document> that contains the added schema.
 
@@ -1614,7 +1633,12 @@ otherwise returns the L<JSON::Schema::Modern::Document> that contains the added
 
 Introduces the L<JSON::Schema::Modern::Document> (or subclass)
 object, representing a JSON Schema, to the implementation, registering it under the indicated URI if
-provided (and all known identifiers within the document will be added as well).
+provided (and all known identifiers within the document will be added as well, resolved against the
+provided URI).
+
+If you add multiple documents (either with this method, or implicitly via L</evaluate>) with no root
+identifier (either provided explicitly in the method call, or via an C<$id> keyword at the schema
+root), all such previous schemas are removed from memory and can no longer be referenced.
 
 If there were errors in the document, will die with these errors;
 otherwise returns the L<JSON::Schema::Modern::Document> object.
